==> rac-1.2-devel1.tgz: lib_rac.c <==
/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"


/* Allocate N bytes of memory dynamically 
 * string[] is used to specify the error message */
void *xmalloc(size_t n, char string[])
{
	void *p=NULL;

	if ((p = malloc(n)) == 0)
		error("xmalloc", "%s", string);

	return p;
}

/* print error messagem, remove (if needed) the temp file
 * and exit 
 *
 * PS: to print error messages without the function name, call 
 * this function like: error("", "message")
 */
void error(char *func, char *fmt, ...)
{
	char msg[MAX_ERROR_SIZE];
	va_list args;

	va_start(args, fmt);
	vsnprintf(msg, sizeof(msg)-1 , fmt, args);
	if (func != "")
		fprintf(stderr, "Function %s, Erro: %s\n",func, msg);
	else
		fprintf(stderr, "Erro: %s\n",msg);
	va_end(args);

	temp_file(REMOVE_TEMP_FILE);
	fflush(stderr);
	exit(1);
}

/* it the locates stream it in the beginning of one specific line
 *
 * to walk in the maximum the half of the vector
 * this is a log(n/2) algorithm, where n is the
 * input's total lines 
 */
void locate_line(long int num_line)
{
	static long int file_pos=0, last_locate_line=-1;
	long int i, j, k;

	/* check if the line is between the range 1 and input's total_lines*/
	if (num_line > rac.total_lines || num_line < 1)
		error("locate_line", "num_line = %li total_lines = %li",num_line,rac.total_lines);
	/* line 'i' is in the num_chars_line vector i-1
	 * first position in vector is zero */
	num_line--;
	
	/* if abs is not portable, change to if (i<0) i =- i; */
	i = abs(0 - num_line);
	j = abs((rac.total_lines-1) - num_line);
	k = abs(last_locate_line - num_line);

	/* test the shorter way to walk 
	 * start from the begin with file_pos = 0 */
	if (i <= j && i <= k) {
		file_pos = 0;
		for (i=0;i<num_line;i++)
			file_pos += rac.num_chars_line[i];
		if (fseek(rac.stream, file_pos, SEEK_SET) != 0)
			error("locate_line", "fseek");
	}
	/* start from the end with file_pos = 0 */
	else if(j <= i && j <= k) {
		file_pos = 0;
		for (i=rac.total_lines-1;i>=num_line;i--)
			file_pos -= rac.num_chars_line[i];
		if (fseek(rac.stream, file_pos, SEEK_END) != 0)
			error("locate_line", "fseek");
	}
	/* the shorter way is from the last_locate_line
	 * so, do not change the file_pos */
	else {
		if (last_locate_line <= num_line)
			for (i=last_locate_line;i<=num_line-1;i++)
				file_pos += rac.num_chars_line[i];
		else
			for (i=last_locate_line-1;i>=num_line;i--) 
				file_pos -= rac.num_chars_line[i];
		/* test if file_pos is positive or negative to
		 * used the proper 'whence', ie, start or end file to 'fseek' */
		if (file_pos < 0) {
			if (fseek(rac.stream, file_pos, SEEK_END) != 0)
				error("locate_line", "fseek");
		}
		else {
			if (fseek(rac.stream, file_pos, SEEK_SET) != 0)
				error("locate_line", "fseek");
		}
	}
	last_locate_line = num_line;
	/* if you do not want to count \n, use num_chars_line[i] less 1.
	 * for debug only. this comment and printf will be erased 
	printf("first char = %c  line size %li\n",getc(stream),num_chars_line[num_line]); */
}

/* read from stream and return a line */
char *get_line(long int line)
{
	char *buf=NULL;

	line--;
	buf = xmalloc(rac.num_chars_line[line] + 1, "get_line: alloc buf.");
	if (fread(buf, sizeof (char), rac.num_chars_line[line], rac.stream) <
					rac.num_chars_line[line] )
		if (!feof(rac.stream))
			error("get_line", "reading input file.");
	buf[rac.num_chars_line[line]] = '\0';
	return buf;
}

/* return the type of address 
 * used to differ the execution 
 * or a error with it is a invalid range 
 */
int get_command_type(char *command)
{
	regex_t *ER;
	int status = -1;

	/* Make space for the regular expression */
	ER = (regex_t *) alloca(sizeof(regex_t));
	memset(ER, 0, sizeof(regex_t));

	/* Compile the regular expression pointed to by pattern.
	 * and check the column range type*/
	if ((regcomp(ER , ER_TYPE_RANGE_COL, REG_EXTENDED | REG_NOSUB)) != 0) 
		error("","regcomp");
	/* try to match the range col ER */
	if ((regexec(ER, command, (size_t)0, (regmatch_t *)NULL, 0)) == 0)
		status = TYPE_COL_RANGE;

	memset(ER, 0, sizeof(regex_t));
	/* Compile the regular expression pointed to by pattern.
	 * and check the line range type */
	if ((regcomp(ER , ER_TYPE_RANGE_LINE, REG_EXTENDED | REG_NOSUB)) != 0) 
		error("","regcomp");
	/* try to match the range line ER */
	if ((regexec(ER, command, (size_t)0, (regmatch_t *)NULL, 0)) == 0)
		status = TYPE_LINE_RANGE;

	if (status == -1) 
		error("","Invalid Range.");
	return status;
}


/* read the command line. 
 * it separates command by command and put it into a matrix 
 * return the commands number 
 */
int read_cmd(char ***command, char *line)
{
	long int i, j, k;
	int cmd_num, cmd_size;
	char *token;

	*command = xmalloc(CMD_SIZE * sizeof (char *),
			"exec_cmd alloc command's memory.");

	cmd_num = CMD_SIZE;
	cmd_size = 0;

	/* Get command by command */
	token = strtok(line, ";");
	k=0;
	while (token != NULL) {
		for (i=0, j=0;i<strlen(token);i++) {
			if (cmd_size == 0) 	(*command)[k] = NULL;
			/* test if we need to realloc (enlarge) the size of this
			 * specific command */
			if (j >= cmd_size - 1) {
#ifdef DEBUG2
				printf("--- line = %li, realloc size of commands = %d\n",k,cmd_size);
#endif
				if (((cmd_size+CMD_SIZE)-INT_MAX) > 0)
					error("read_cmd", "overflow cmd_size");

				cmd_size += CMD_SIZE;
				(*command)[k] = (char *) realloc ((*command)[k], cmd_size * sizeof (char));
				if (!(*command)[k])
					error("read_cmd", "realloc command's memory");
			}
			/* erase the blank characters */
			if (token[i] != ' ' && token[i] != '\t') (*command)[k][j++] = token[i];
		}
		token = strtok(NULL, ";");
		(*command)[k][j] = '\0';
		cmd_size = 0;
		k++;
		/* test if we need to realloc (enlarge) the number of commands 
		 * command[k] */
		if (k >= cmd_num) {
#ifdef DEBUG2
			printf("--- realloc number of commands = %d\n",cmd_num);
#endif
			if (((cmd_num+CMD_SIZE)-INT_MAX) > 0)
				error("read_cmd", "overflow cmd_num");

			cmd_num += CMD_SIZE;
			*command = (char **) realloc(*command, cmd_num * sizeof(char **));
			if (!*command)
				error("read_cmd", "realloc command's memory");
		}
	}
	/* return the number of commands
	 * if it has one command and nothing in the command[k], so it returns 0
	 * ex: ./rac '      ' file */
	return (k == 1 && (*command)[k-1][0] == '\0' ? 0 : k);
}


/* extract the strings inside double quotes
 * just from line type addressing directives
 */
void extract_strings(char *addrs)
{
	char col_mode = 0, str_mode = 0;
	char *str_start = NULL;
	struct address_string *str;

	while (*addrs) {
		switch (*addrs) {
		case '\\':
			if (*(addrs + 1)) addrs++;
			break;
		case '[':
			if (!str_mode) col_mode = 1;
			break;
		case ']':
			if (!str_mode) col_mode = 0;
			break;
		case '=':
			if (!col_mode) {
				if (str_mode) {
					*addrs = '\0';
					/* search string in rac.strings */
					str = rac.strings;
					while (str) {
						if (!strcmp(str_start, str->string)) break;
						str = str->next;
					}
					/* add string in rac.strings if it is not there */
					if (!str) {
						str = xmalloc(sizeof(struct address_string), "extract_strings alloc address_string.");
						str->next = rac.strings;
						rac.strings = str;
						str->string = xmalloc(strlen(str_start)+1, "extract_strings alloc str->string.");
						str->escaped = xmalloc(strlen(str_start)+1, "extract_strings alloc str->escaped.");
						strcpy(str->string, str_start);
						strcpy(str->escaped, str_start);

						escape_string(str->escaped);

						str->occurs = NULL;
						str->num_occurs = 0;
					}
					*addrs = '=';
					str_mode = 0;
				}
				else {
					str_start = addrs + 1;
					str_mode = 1;
				}
			}
			break;
		}
		addrs++;
	}
	if (str_mode || col_mode)
		error("", "Invalid addressing.");
}

/* interpret the backslash-escaped characters: \\ \= \t \n \r
 * edit string in place
 */
void escape_string(char *str)
{
	int i, j;

	for (i = 0, j = 0; str[i]; i++, j++) {
		if (str[i] == '\\' && str[i+1]) {
			i++;
			switch (str[i]) {
			case '\\': break;
			case '=': str[j] = '='; break;
			case 't': str[j] = '\t'; break;
			case 'n': str[j] = '\n'; break;
			case 'r': str[j] = '\r'; break;
			}
		}
		else 
			str[j] = str[i];
	}
	str[j] = '\0';
}

==> rac-1.3-devel1.tgz: lib_rac.c <==
/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"


/* Allocate N bytes of memory dynamically 
 * string[] is used to specify the error message */
void *xmalloc(size_t n, char string[])
{
	void *p=NULL;

	if ((p = malloc(n)) == 0)
		error("xmalloc", "%s", string);

	return p;
}

/* print error messagem and exit 
 *
 * PS: to print error messages without the function name, call 
 * this function like: error("", "message")
 */
void error(char *func, char *fmt, ...)
{
	char msg[MAX_ERROR_SIZE];
	va_list args;

	va_start(args, fmt);
	vsnprintf(msg, sizeof(msg)-1 , fmt, args);
	if (func != "")
		fprintf(stderr, "Function %s, Erro: %s\n",func, msg);
	else
		fprintf(stderr, "Erro: %s\n",msg);
	va_end(args);

	fflush(stderr);
	exit(1);
}

/* it the locates stream it in the beginning of one specific line
 *
 * to walk in the maximum the half of the vector
 * this is a log(n/2) algorithm, where n is the
 * input's total lines 
 */
void locate_line(long int num_line)
{
	static long int file_pos=0, last_locate_line=-1;
	long int i, j, k;

	/* check if the line is between the range 1 and input's total_lines*/
	if (num_line > rac.total_lines || num_line < 1)
		error("locate_line", "num_line = %li total_lines = %li",num_line,rac.total_lines);
	/* line 'i' is in the num_chars_line vector i-1
	 * first position in vector is zero */
	num_line--;
	
	/* if abs is not portable, change to if (i<0) i =- i; */
	i = abs(0 - num_line);
	j = abs((rac.total_lines-1) - num_line);
	k = abs(last_locate_line - num_line);

	/* test the shorter way to walk 
	 * start from the begin with file_pos = 0 */
	if (i <= j && i <= k) {
		file_pos = 0;
		for (i=0;i<num_line;i++)
			file_pos += rac.num_chars_line[i];
		if (fseek(rac.stream, file_pos, SEEK_SET) != 0)
			error("locate_line", "fseek");
	}
	/* start from the end with file_pos = 0 */
	else if(j <= i && j <= k) {
		file_pos = 0;
		for (i=rac.total_lines-1;i>=num_line;i--)
			file_pos -= rac.num_chars_line[i];
		if (fseek(rac.stream, file_pos, SEEK_END) != 0)
			error("locate_line", "fseek");
	}
	/* the shorter way is from the last_locate_line
	 * so, do not change the file_pos */
	else {
		if (last_locate_line <= num_line)
			for (i=last_locate_line;i<=num_line-1;i++)
				file_pos += rac.num_chars_line[i];
		else
			for (i=last_locate_line-1;i>=num_line;i--) 
				file_pos -= rac.num_chars_line[i];
		/* test if file_pos is positive or negative to
		 * used the proper 'whence', ie, start or end file to 'fseek' */
		if (file_pos < 0) {
			if (fseek(rac.stream, file_pos, SEEK_END) != 0)
				error("locate_line", "fseek");
		}
		else {
			if (fseek(rac.stream, file_pos, SEEK_SET) != 0)
				error("locate_line", "fseek");
		}
	}
	last_locate_line = num_line;
	/* if you do not want to count \n, use num_chars_line[i] less 1.
	 * for debug only. this comment and printf will be erased 
	printf("first char = %c  line size %li\n",getc(stream),num_chars_line[num_line]); */
}

/* read from stream and return a line */
char *get_line(long int line)
{
	char *buf=NULL;

	line--;
	buf = xmalloc(rac.num_chars_line[line] + 1, "get_line: alloc buf.");
	if (fread(buf, sizeof (char), rac.num_chars_line[line], rac.stream) <
					rac.num_chars_line[line] )
		if (!feof(rac.stream))
			error("get_line", "reading input file.");
	buf[rac.num_chars_line[line]] = '\0';
	return buf;
}

/* return the type of address 
 * used to differ the execution 
 * or a error with it is a invalid range 
 */
int get_command_type(char *command)
{
	regex_t *ER;
	int status = -1;

	/* Make space for the regular expression */
	ER = (regex_t *) alloca(sizeof(regex_t));
	memset(ER, 0, sizeof(regex_t));

	/* Compile the regular expression pointed to by pattern.
	 * and check the column range type*/
	if ((regcomp(ER , ER_TYPE_RANGE_COL, REG_EXTENDED | REG_NOSUB)) != 0) 
		error("","regcomp");
	/* try to match the range col ER */
	if ((regexec(ER, command, (size_t)0, (regmatch_t *)NULL, 0)) == 0)
		status = TYPE_COL_RANGE;

	memset(ER, 0, sizeof(regex_t));
	/* Compile the regular expression pointed to by pattern.
	 * and check the line range type */
	if ((regcomp(ER , ER_TYPE_RANGE_LINE, REG_EXTENDED | REG_NOSUB)) != 0) 
		error("","regcomp");
	/* try to match the range line ER */
	if ((regexec(ER, command, (size_t)0, (regmatch_t *)NULL, 0)) == 0)
		status = TYPE_LINE_RANGE;

	if (status == -1) 
		error("","Invalid Range.");
	return status;
}


/* read the command line. 
 * it separates command by command and put it into a matrix 
 * return the commands number 
 */
int read_cmd(char ***command, char *line)
{
	long int i, j, k;
	int cmd_num, cmd_size;
	char *token;

	*command = xmalloc(CMD_SIZE * sizeof (char *),
			"exec_cmd alloc command's memory.");

	cmd_num = CMD_SIZE;
	cmd_size = 0;

	/* Get command by command */
	token = strtok(line, ";");
	k=0;
	while (token != NULL) {
		int n = strlen(token);
		for (i=0, j=0; i < n; i++) {
			if (cmd_size == 0) 	(*command)[k] = NULL;
			/* test if we need to realloc (enlarge) the size of this
			 * specific command */
			if (j >= cmd_size - 1) {
#ifdef DEBUG2
				printf("--- line = %li, realloc size of commands = %d\n",k,cmd_size);
#endif
				if (((cmd_size+CMD_SIZE)-INT_MAX) > 0)
					error("read_cmd", "overflow cmd_size");

				cmd_size += CMD_SIZE;
				(*command)[k] = (char *) realloc ((*command)[k], cmd_size * sizeof (char));
				if (!(*command)[k])
					error("read_cmd", "realloc command's memory");
			}
			/* erase the blank characters */
			if (token[i] != ' ' && token[i] != '\t') (*command)[k][j++] = token[i];
		}
		token = strtok(NULL, ";");
		(*command)[k][j] = '\0';
		cmd_size = 0;
		k++;
		/* test if we need to realloc (enlarge) the number of commands 
		 * command[k] */
		if (k >= cmd_num) {
#ifdef DEBUG2
			printf("--- realloc number of commands = %d\n",cmd_num);
#endif
			if (((cmd_num+CMD_SIZE)-INT_MAX) > 0)
				error("read_cmd", "overflow cmd_num");

			cmd_num += CMD_SIZE;
			*command = (char **) realloc(*command, cmd_num * sizeof(char **));
			if (!*command)
				error("read_cmd", "realloc command's memory");
		}
	}
	/* return the number of commands
	 * if it has one command and nothing in the command[k], so it returns 0
	 * ex: ./rac '      ' file */
	return (k == 1 && (*command)[k-1][0] == '\0' ? 0 : k);
}


/* extract the strings inside = = and the RE's inside / /
 * only from line type addressing directives
 */
void extract_strings(char *addrs)
{
	char col_mode = 0, str_mode = 0, re_mode = 0;
	char *str_start = NULL;

	while (*addrs) {
		switch (*addrs) {
		case '\\':
			if (*(addrs + 1)) addrs++;
			break;
		case '[':
			if (!str_mode && !re_mode) col_mode = 1;
			break;
		case ']':
			if (!str_mode && !re_mode) col_mode = 0;
			break;
		case '=':
			if (!col_mode && !re_mode) {
				if (str_mode) {
					*addrs = '\0';
					add_string(str_start, 0);
					*addrs = '=';
					str_mode = 0;
				}
				else {
					str_start = addrs + 1;
					str_mode = 1;
				}
			}
			break;
		case '/':
			if (!col_mode && !str_mode) {
				if (re_mode) {
					*addrs = '\0';
					add_string(str_start, 1);
					*addrs = '/';
					re_mode = 0;
				}
				else {
					str_start = addrs + 1;
					re_mode = 1;
				}
			}
			break;
		}
		addrs++;
	}
	if (str_mode || col_mode || re_mode)
		error("", "Invalid addressing.");
}


/* interpret the backslash-escaped characters: \t \n \r
 * don't interpret \\ in RE's
 * delim is '=' (to string) or '/' (to RE)
 * edit str in place
 */
void escape_string(char *str, char delim)
{
	int i, j;

	for (i = 0, j = 0; str[i]; i++, j++) {
		if (str[i] == '\\' && str[i+1]) {
			i++;
			if (str[i] == delim) str[j] = delim;
			else
			  switch (str[i]) {
			  case '\\':
			    /* keep double backslash in RE */
			    if (delim == '/') str[j++] = '\\';
			    str[j] = '\\';
			    break;
			  case 't': str[j] = '\t'; break;
			  case 'n': str[j] = '\n'; break;
			  case 'r': str[j] = '\r'; break;
			  default:
			    /* keep backslash if is an unknown escape */
			    str[j++] = '\\';
			    str[j] = str[i];
			  }
		}
		else 
			str[j] = str[i];
	}
	str[j] = '\0';
}


/* add string or regular expression in rac.strings
 */
void add_string(char *str, char isRE)
{
  struct address_string *rs;
  
  if (!str || !*str) error("", "Invalid addressing.");

  /* search string in rac.strings */
  rs = rac.strings;
  while (rs) {
    if (rs->isRE == isRE && !strcmp(str, rs->string)) break;
    rs = rs->next;
  }
  /* add string in rac.strings if it is not there */
  if (!rs) {
    rs = xmalloc(sizeof(struct address_string), "add_string alloc address_string.");
    rs->next = rac.strings;
    rac.strings = rs;
    rs->string = xmalloc(strlen(str)+1, "add_string alloc rs->string.");
    rs->escaped = xmalloc(strlen(str)+1, "add_string alloc rs->escaped.");
    strcpy(rs->string, str);
    strcpy(rs->escaped, str);
    
    escape_string(rs->escaped, (isRE ? '/' : '='));
    
    rs->isRE = isRE;
    rs->occurs = NULL;
    rs->num_occurs = 0;
  }
}
